//
// CSE223B Lab 3: Distributed, Multi-party Chat Server
//
// Copyright(c) 2003-6 Alex C. Snoeren
//

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <list>
#include <string>
#include <iostream>

using namespace std;

#define MAXCHATLINE 255


class Client;

class EventHandler {
public:
  virtual int readHandler() = 0;
  virtual int writeHandler() = 0;
};

class ChatServer : public EventHandler {

private:

  int _id;
  int _s;

  list<struct sockaddr_in> _peers;
  list<Client *> _clients;

public:

  ChatServer(int id, int port);
  int addPeer(const char *, int);
  int addClient(Client * c);
  int removeClient(Client * c);
  int getSocket(void) { return _s; }
  int dispatchMessage(string);
  int readHandler() {}
  int writeHandler() {} // We won't need this

};


ChatServer::ChatServer(int id, int port)
{
  _id = id;

  // new socket
  if((_s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    perror("socket");
    exit(1);
  }

  // don't whine about address already in use
  int n = 1;
  if(setsockopt(_s, SOL_SOCKET, SO_REUSEADDR, (char *)&n, sizeof(n)) < 0) {
    perror("setsockopt");
    exit(1);
  }

  // bind socket to some port
  sockaddr_in sin;
  bzero(&sin, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons(port);
  sin.sin_addr.s_addr = INADDR_ANY;
  if(bind(_s, (sockaddr*) &sin, sizeof(sin)) < 0) {
    perror("bind");
    exit(1);
  }
}

int
ChatServer::addClient(Client * c)
{
  _clients.push_front(c);
}

int
ChatServer::removeClient(Client * c)
{
  _clients.remove(c);
}

int
ChatServer::addPeer(const char *hostname, int port)
{
  struct hostent *hp;
  if(!(hp = gethostbyname(hostname))) {
    cerr << "Unknown host: " << hostname << endl;
    return -1;
  }

  sockaddr_in sin;
  bzero(&sin, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons(port);
  sin.sin_addr.s_addr = *(in_addr_t *)hp->h_addr;

  _peers.push_front(sin);
  return 0;
}


EventHandler * connHandlers[FD_SETSIZE];
fd_set rfds, wfds;
ChatServer * cs;

class Client : public EventHandler {

private:
  // Keep peer-connection state here
  int _fd;
  string clientName;
  string _buffer;

public:

  Client(int fd) : _fd(fd), clientName(""), _buffer("") {}
  ~Client();
  int readHandler();
  int writeHandler();
  int sendMessage(string message);
  string getClientName() { return clientName; }
};

Client::~Client()
{
  cs->removeClient(this);
  connHandlers[_fd] = NULL;
  FD_CLR(_fd, &rfds);
  FD_CLR(_fd, &wfds);
  close(_fd);
}

int
ChatServer::dispatchMessage(string message)
{
  // XXX: You need to extend this to send the message to other peer servers
  for(list<Client *>::iterator i =_clients.begin(); i!=_clients.end();++i) {
    (*i)->sendMessage(message);
  }
}

// This is a hack--we assume we never receive less than a line at a time.
int
readln(int fd, char *buf, int len)
{
  int n, i = 0;
  char c;
  while(i < len) {
    if((n = read(fd, &buf[i], 1)) < 0) {
      perror("readln");
      exit(1);
    }
    if(n == 0)
      return 0;
    if(buf[i] == '\n') {
      buf[++i] = '\0';
      break;
    }
    if(buf[i] == '\0')
      break;
    ++i;
  }
  return i;
}

int
Client::readHandler()
{
  char buf[MAXCHATLINE];
  ssize_t nread = readln(_fd, buf, MAXCHATLINE);

  // If it's disconnected, terminate nicely
  if(!nread) {
    cout << "Client " << clientName << " disconnected." << endl;
    delete this;
    return 1;
  }

  // If we haven't yet got a name, get one.
  if(clientName == "") {
    clientName = buf;
    cout << "Client " << clientName << " connected." << endl;
  } else {
    string message = clientName + ": " + string(buf) + '\n';
    // Send the message to everyone
    cs->dispatchMessage(message);
  }
  return 0;
}

int
Client::writeHandler()
{
  ssize_t n = write(_fd, _buffer.c_str(), _buffer.size());
  if(n < 0) {
    perror("write");
    exit(1);
  }
  _buffer.erase(0, n);
  if(!_buffer.size())
    FD_CLR(_fd, &wfds);
}

int
Client::sendMessage(string message)
{
  _buffer.append(message);
  FD_SET(_fd, &wfds);
}

int
setupClientPort(int port)
{
  int s = -1;

  // new socket
  if((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    perror("socket");
    return -1;
  }

  // don't whine about address already in use
  int n = 1;
  if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&n, sizeof(n)) < 0) {
    perror("setsockopt");
    return -1;
  }

  // bind socket to some port
  sockaddr_in sin;
  bzero(&sin, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons(port);
  sin.sin_addr.s_addr = INADDR_ANY;
  if(bind(s, (sockaddr*) &sin, sizeof(sin)) < 0) {
    perror("bind");
    exit(1);
  }

  // listen for incoming connections
  if(listen(s, 5) < 0) {
    perror("listen");
    exit(1);
  }

  return s;
}


int
acceptConn(int fd)
{
  sockaddr_in sin;
  socklen_t sinlen = sizeof(sin);
  int c;

  if((c = accept(fd, (sockaddr*) &sin, &sinlen)) < 0) {
    perror("accept");
    return -1;
  }

  Client * cl = new Client(c);
  connHandlers[c] = cl;
  cs->addClient(cl);

  // Reister interest in a read on this socket
  FD_SET(c, &rfds);

  return(c);
}

int
main(int argc, char *argv[])
{
  int c = -1;

  if((argc < 3) || !(argc % 2)) {
    cerr << "Usage: " << argv[0] << " id localport peer1 port1..." << endl;
    exit(1);
  }

  cs = new ChatServer(atoi(argv[1]), atoi(argv[2]));
  connHandlers[cs->getSocket()] = cs;

  for(int i = 3; i+1 < argc; i += 2) {
    if(cs->addPeer(argv[i], atoi(argv[i+1])))
      exit(1);
  }

  // Prepare to accept clients
  FD_ZERO(&rfds);
  FD_ZERO(&wfds);
  if((c = setupClientPort(atoi(argv[2]))) < 0)
    exit(1);
  FD_SET(cs->getSocket(), &rfds);
  FD_SET(c, &rfds);

  // Event loop
  while(1) {

    int n = 0;
    fd_set _rfds, _wfds;

    memcpy(&_rfds, &rfds, sizeof(fd_set));
    memcpy(&_wfds, &wfds, sizeof(fd_set));

    // Block until there's an event to handle
    if(((n = select(FD_SETSIZE, &_rfds, &_wfds, NULL, NULL)) < 0)) {
      perror("select");
      exit(1);
    }

    // Handle events
    for(int i = 0; (i < FD_SETSIZE) && n ; i++) {

      if(FD_ISSET(i, &_rfds)) {
	// Handle our listening socket differently
	if(i == c) {
	  if(acceptConn(c) < 0)
	    exit(1);
	} else {
	  if(connHandlers[i]) {
	    if(connHandlers[i]->readHandler()) {
	      FD_CLR(i, &rfds);
	      FD_CLR(i, &wfds);
	    }
	  }
	}
	n--;
      }
      if(FD_ISSET(i, &_wfds)) {
	if(connHandlers[i])
	  connHandlers[i]->writeHandler();
	n--;
      } 
    }
  }
}
